"""Tools for manipulating electrophysiological data."""

from . import pltools
from . import rectools
from . import stimtools

__author__ = 'Emerson Harkin'
__email__ = 'emerson@efharkin.com'
__version__ = '0.4.3'



# IMPORT MODULES

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

# DEFINE HANDY TOOLS

# Function to make a set of scalebars for a mpl plot


def add_scalebar(x_units=None, y_units=None, anchor=(0.98, 0.02),
                 x_size=None, y_size=None, y_label_space=0.02, x_label_space=-0.02,
                 bar_space=0.06, x_on_left=True, linewidth=3, remove_frame=True,
                 omit_x=False, omit_y=False, round=True, usetex=True, ax=None):
    """Automagically add a set of x and y scalebars to a matplotlib plot.

    By default, scalebars are sized automatically based on x and y tick
    spacing.

    Arguments
    ---------
    x_units, y_units: str or None
        Units (e.g. mV, pA, ms) for x and y scalebars.
    anchor: pair of floats
        Bottom right of scale bar bbox (in axis coordinates).
    x_size, y_size: float or None
        Manually set size of x and y scalebars (or None for automatic sizing).
    x_label_space, y_label_space: float
        Offset for x and y scalebar labels (in axis units).
    bar_space: float >= 0.0
        Spacing between x and y scalebars (in axis units). Set to zero for the
        two bars to be joined at the corner.
    x_on_left: bool, default True
        Place x scalebar on left of y scalebar. Set to False to place on right
        instead.
    linewidth: float, default 3
        Thickness of scalebars.
    remove_frame: bool, default True
        Remove the frame, ticks, etc. from the axis object to which the
        scalebar is being added. True by default because the scalebars are
        assumed to be a replacement for the axes.
    omit_x, omit_y: bool, default False
        Skip drawing x or y scalebar.f
    round: bool, default True
        Round values to the nearest integer. Causes `100.0 pA` to be printed as
        `100 pA`.
    usetex: bool, default True
        Print numbers in LaTeX math mode. Set to False if LaTeX is not being
        used to render text, or `100 pA` will print as `$100$ pA`.
    ax: matplotlib.axes object, defaults to current axes
        Axis to which scalebars should be added.

    """
    # Basic input processing.

    if ax is None:
        ax = plt.gca()

    if x_units is None:
        x_units = ''
    if y_units is None:
        y_units = ''

    # Do y scalebar.
    if not omit_y:

        if y_size is None:
            y_span = ax.get_yticks()[:2]
            y_length = y_span[1] - y_span[0]
            y_span_ax = ax.transLimits.transform(
                np.array([[0, 0], y_span]).T)[:, 1]
        else:
            y_length = y_size
            y_span_ax = ax.transLimits.transform(
                np.array([[0, 0], [0, y_size]]))[:, 1]
        y_length_ax = y_span_ax[1] - y_span_ax[0]

        if round:
            y_length = int(np.round(y_length))

        # y-scalebar label

        if y_label_space <= 0:
            horizontalalignment = 'left'
        else:
            horizontalalignment = 'right'

        if usetex:
            y_label_text = '${}${}'.format(y_length, y_units)
        else:
            y_label_text = '{}{}'.format(y_length, y_units)

        ax.text(
            anchor[0] - y_label_space, anchor[1] + y_length_ax / 2 + bar_space,
            y_label_text,
            verticalalignment='center', horizontalalignment=horizontalalignment,
            size='small', transform=ax.transAxes
        )

        # y scalebar
        ax.plot(
            [anchor[0], anchor[0]],
            [anchor[1] + bar_space, anchor[1] + y_length_ax + bar_space],
            'k-',
            linewidth=linewidth,
            clip_on=False,
            solid_capstyle='butt',
            transform=ax.transAxes
        )

    # Do x scalebar.
    if not omit_x:

        if x_size is None:
            x_span = ax.get_xticks()[:2]
            x_length = x_span[1] - x_span[0]
            x_span_ax = ax.transLimits.transform(
                np.array([x_span, [0, 0]]).T)[:, 0]
        else:
            x_length = x_size
            x_span_ax = ax.transLimits.transform(
                np.array([[0, 0], [x_size, 0]]))[:, 0]
        x_length_ax = x_span_ax[1] - x_span_ax[0]

        if round:
            x_length = int(np.round(x_length))

        # x-scalebar label
        if x_label_space <= 0:
            verticalalignment = 'top'
        else:
            verticalalignment = 'bottom'

        if x_on_left:
            Xx_text_coord = anchor[0] - x_length_ax / 2 - bar_space
            Xx_bar_coords = [
                anchor[0] - x_length_ax - bar_space,
                anchor[0] - bar_space]
        else:
            Xx_text_coord = anchor[0] + x_length_ax / 2 + bar_space
            Xx_bar_coords = [
                anchor[0] + x_length_ax + bar_space,
                anchor[0] + bar_space]

        if usetex:
            x_label_text = '${}${}'.format(x_length, x_units)
        else:
            x_label_text = '{}{}'.format(x_length, x_units)

        ax.text(
            Xx_text_coord, anchor[1] + x_label_space,
            x_label_text,
            verticalalignment=verticalalignment, horizontalalignment='center',
            size='small', transform=ax.transAxes
        )

        # x scalebar
        ax.plot(
            Xx_bar_coords,
            [anchor[1], anchor[1]],
            'k-',
            linewidth=linewidth,
            clip_on=False,
            solid_capstyle='butt',
            transform=ax.transAxes
        )

    if remove_frame:
        ax.axis('off')


def hide_border(sides='a', trim=False, ax=None):
    """Remove and/or trim axes borders.

    Most common use is to remove top and right border from matplotlib axes.

    Arguments
    ---------
    sides : str `a`, one or more of `rltb`, or `none`
        Borders to remove. Use `a` for all, r` for right border, `rl` for right
        and left, etc.
    trim : bool, default False
        Shorten remaining axes to first and last tick. See `seborn.despine`.
    ax : matplotlib.axes object, defaults to current axes
        Hide border of this axes object.

    """
    # Check for correct input
    if not any([letter in sides for letter in 'arltb']):
        raise ValueError(
            'sides should be passed a string with `a` for all sides, '
            'or r/l/t/b as-needed for other sides.')

    if ax is None:
        ax = plt.gca()

    if sides == 'a':
        sides = 'rltb'

    # Remove ticks if needed.
    if 'l' in sides:
        ax.set_yticks([])
    if 'b' in sides:
        ax.set_xticks([])

    # Remove border(s); wraps seaborn.despine().
    sidekeys = {
        'r': 'right',
        'l': 'left',
        't': 'top',
        'b': 'bottom'
    }
    snsdespine_side_args = {}
    for key in sidekeys:
        if key in sides:
            snsdespine_side_args[sidekeys[key]] = True
        else:
            snsdespine_side_args[sidekeys[key]] = False
    sns.despine(trim=trim, ax=ax, **snsdespine_side_args)


def hide_ticks(ax=None):
    """Remove x and y ticks from axes object.

    Arguments
    ---------
    ax: matplotlib.axes, defaults to current axes

    """
    if ax is None:
        ax = plt.gca()

    ax.set_xticks([])
    ax.set_yticks([])


def p_to_string(p):
    """Convert p-value to pretty latex string.

    p is presented to three decimal places if p >= 0.05, and as
    p < 0.05/0.01/0.001 otherwise.

    Arguments
    ---------
    p: float
        p-value to format.

    Returns
    -------
        p-value formatted as a string.

    """
    p_rounded = np.round(p, 3)

    if p_rounded >= 0.05:
        p_str = '$p = {}$'.format(p_rounded)
    elif p_rounded < 0.05 and p_rounded >= 0.01:
        p_str = '$p < 0.05$'
    elif p_rounded < 0.01 and p_rounded >= 0.001:
        p_str = '$p < 0.01$'
    else:
        p_str = '$p < 0.001$'

    return p_str



# IMPORT MODULES

import os

import matplotlib.pyplot as plt

from ezephys.rectools import ABFLoader
from ezephys import pltools


# LOAD TEST RECORDING

TEST_REC_PATH = os.path.join('test_data', '17n28000.abf')
loader = ABFLoader()
test_rec = loader.load(TEST_REC_PATH)[0]
test_rec.plot()

# TRY ADDING SCALEBAR

plt.figure(figsize=(10, 5))

plt.subplot2grid((4, 1), (0, 0), rowspan=3)
plt.plot(test_rec[0, :, :].mean(axis=1), 'k-')

plt.xlim(1000, 8000)
plt.ylim(-200, 200)
pltools.add_scalebar(y_units='pA', x_units='ms', anchor=(0.98, 0.2))

plt.subplot2grid((4, 1), (3, 0))
plt.plot(test_rec[1, :, :].mean(axis=1), 'k-')

plt.xlim(1000, 8000)

pltools.add_scalebar(y_units='mV', x_units='ms', anchor=(0.7, 0.15),
                     y_label_space=0.007, bar_space=0.03, omit_x=True)

plt.show()



"""Construct complex electrophysiological stimuli.

@author: Emerson

"""
# IMPORT MODULES

import csv
import warnings
from copy import deepcopy

import numpy as np
import numba as nb
import matplotlib.pyplot as plt


# SIMULUS BASE CLASS

class BaseStimulus(object):
    """Abstract base class for stimulus objects.

    Attributes
    ----------
    no_sweeps: int
        Number of sweeps in stimulus.
    no_timesteps: int
        Number of timesteps in one stimulus sweep.
    duration: float
        Duration of one stimulus sweep in ms.
    dt: float
        Timestep of stimulus in ms.
    label: str
        Descriptive label.

    Methods
    -------
    generate
        Generate a discrete representation of the stimulus. Useful for
        visualization and export.
    replicate(replicates)
        Add sweeps to the stimulus by replicating a single sweep.
    copy
    simulate_response(R, C, E, plot, verbose)
        Simulate the voltage response of a neuron to the stimulus.
    plot
        Create a matplotlib plot to visualize the stimulus.
    export
        Export the stimulus for use in ephys experiments.

    """

    # Methods that must be implemented by derived classes. (Pure virtual.)
    def __init__(self):
        """Initialize Stimulus (implemented by derived classes)."""
        raise NotImplementedError(
            'Initialization must be implemented by derived stimulus classes.'
        )

    # Methods that should not be changed by derived classes.
    def __add__(self, x):
        """Add x to Stimulus object."""
        # Adding is implemented for CompoundStimulus objects,
        # so coerce self to a CompoundStimulus and add.
        tmpstimulus = CompoundStimulus(self)
        return tmpstimulus + x

    def __sub__(self, x):
        """Subtract x from Stimulus object."""
        # Subtraction is implemented for CompoundStimulus objects,
        # so coerce self to a CompoundStimulus and subtract.
        tmpstimulus = CompoundStimulus(self)
        return tmpstimulus - x

    @property
    def no_sweeps(self):
        """Number of stimulus sweeps."""
        if not hasattr(self, 'command'):
            return 0
        elif self.command.ndim == 1:
            return 1
        else:
            assert self.command.ndim == 2
            return self.command.shape[0]

    @property
    def no_timesteps(self):
        """Number of stimulus timesteps."""
        if not hasattr(self, 'command'):
            return 0
        elif self.command.ndim == 1:
            return len(self.command)
        else:
            assert self.command.ndim == 2
            return self.command.shape[1]

    @property
    def duration(self):
        """Duration of stimulus in ms."""
        if not hasattr(self, 'dt'):
            raise AttributeError(
                '`duration` not defined for un-generated stimulus.'
            )
        return self.no_timesteps * self.dt

    def replicate(self, replicates):
        """Replicate stimulus for `replicates` sweeps.

        `replicates=2` results in a two sweep stimulus. `replicates=1` does
        nothing.
        """
        if replicates <= 1:
            raise ValueError('Replicates must be int >1.')
        self.command = np.tile(self.command, (replicates, 1))

    def copy(self):
        """Return deep copy of stimulus instance."""
        return deepcopy(self)

    def simulate_response(self, R, C, E, plot=True, verbose=True):
        """Simulate response of passive membrane to stimulus.

        Useful to see how a neuron recorded in current clamp would be expected
        to respond to the stimulus.

        Models the neuronal membrane as an RC circuit.

        Inputs
        ------
        R: float
            Membrane resistance in MOhm.
        C: float
            Membrane capacitance in pF.
        E: float
            Membrane resting potential in mV.
        plot: bool (default True)
            Plot the simulated voltage response.
        verbose: bool (default True)
            Print information about progress.

        Returns
        -------
        [sweeps, time] array with simulated voltage response.

        """
        # Check that command to export actually exists.
        if not (hasattr(self, 'command') and hasattr(self, 'time_supp')):
            raise AttributeError('`Stimulus<type>.command` must be initialized '
                                 'by calling `Stimulus<type>.generate()` '
                                 'before it can be integrated.')

        # Data type to use for simulations.
        # Needed for numba.jit accelerated current integrator.
        dtype = np.float64

        # Put command into 2D array if it isn't already.
        if self.command.ndim == 1:
            input_ = self.command.copy()[np.newaxis, :].astype(dtype)
        else:
            input_ = self.command.copy().astype(dtype)

        # Unit conversions.
        input_ *= 1e-3
        leak_conductance = 1 / R
        C *= 1e-3
        if verbose:
            print('tau = {}ms'.format(R * C))

        # Run current integrator.
        if verbose:
            print('Integrating input...')
        output_ = np.empty_like(input_)  # Buffer for integrated voltage.
        V = self._integrate(
            input_, output_,
            dtype(leak_conductance), dtype(C), dtype(E), dtype(self.dt)
        )
        #V *= 1e3
        if verbose:
            print('Done!')

        # Plot integrated voltage.
        if plot:
            if verbose:
                print('Plotting...')
            plt.figure()

            ax = plt.subplot(211)
            plt.plot(self.time_supp, V.T, 'k-')
            plt.ylabel('Voltage (mV)')
            plt.xlabel('Time (ms)')

            plt.subplot(212, sharex=ax)
            plt.plot(self.time_supp, self.command.T, 'k-')
            plt.ylabel('Command (pA)')
            plt.xlabel('Time (ms)')

            plt.show()

            if verbose:
                print('Done!')

        return V

    def plot(self, ax=None, **pltargs):
        """Show generated stimulus."""
        # Check that command to plot actually exists.
        if not (hasattr(self, 'command') and hasattr(self, 'time_supp')):
            raise AttributeError('`Stimulus<type>.command` must be initialized '
                                 'by calling `Stimulus<type>.generate()` '
                                 'before it can be plotted.')

        # Get default mpl axes.
        if ax is None:
            ax = plt.gca()

        # Make plot.
        ax.plot(self.time_supp, self.command.T, **pltargs)
        ax.set_xlabel('Time (ms)')
        ax.set_ylabel('Command')

        return ax

    def export(self, fname):
        """Export generated stimulus in Axon text format.

        Compatible with Axon Instruments software.

        Arguments
        ---------
        fname: str
            Name of file to which stimulus will be written.

        """
        # Check that command to export actually exists.
        if not (hasattr(self, 'command') and hasattr(self, 'time_supp')):
            raise AttributeError('`Stimulus<type>.command` must be initialized '
                                 'by calling `Stimulus<type>.generate()` '
                                 'before it can be exported.')

        # Coerce command and time support to column vectors, if applicable.
        arrs_to_export = {}
        for attr in ['command', 'time_supp']:
            if getattr(self, attr).ndim == 1:
                arrs_to_export[attr] = getattr(self, attr)[:, np.newaxis]
            else:
                arrs_to_export[attr] = getattr(self, attr).T
        # Join into a single array to write to file.
        arr_to_export = np.concatenate(
            [arrs_to_export['time_supp'], arrs_to_export['command']],
            axis=1
        )
        del arrs_to_export

        with open(fname, 'w', newline='') as f:
            # Write header.
            f.write(
                'ATF1.0\n1\t{}\nType=1\nTime (ms)\t{}\n'.format(
                    arr_to_export.shape[1] - 1,
                    'Command (AU)\t' * (arr_to_export.shape[1] - 1)
                )
            )

            # Write input.
            stimwriter = csv.writer(f, delimiter='\t')
            stimwriter.writerows(arr_to_export)

            # Close file.
            f.close()

    # Accelerated methods.

    @staticmethod
    @nb.jit(nb.float64[:, :](nb.float64[:, :], nb.float64[:, :], nb.float64,
                             nb.float64, nb.float64, nb.float64), nopython=True)
    def _integrate(input_, output_, leak_conductance, C, E, dt):
        """Integrate `input_` with a passive membrane filter.

        `output_` is a buffer to hold output voltage. Accelerated using
        `numba.jit`.

        """
        for i in range(input_.shape[0]):
            output_[0, i] = E
            for t in range(1, input_.shape[1]):
                dV = (
                    -leak_conductance * (output_[i, t - 1] - E) + input_[i, t - 1]
                ) * dt / C
                output_[i, t] = output_[i, t - 1] + dV

        return output_


# ARBITRARY ARRAY STIMULUS

class ArrayStimulus(BaseStimulus):
    """Stimulus constructed from an arbitrary array.

    See documentation of BaseStimulus for more information about Stimulus
    objects.

    """

    def __init__(self, command, dt=0.1, label=None):
        """Initialize ArrayStimulus.

        Arguments
        ---------
        command: 1D or 2D array
            Stimulus waveform.
        dt: float, default 0.1
            Timestep in ms.
        label: str
            Descriprive label for stimulus instance.

        """
        self.label = label

        if issubclass(type(command), BaseStimulus):
            self.command = command.command
        elif issubclass(type(command), StimulusKernel):
            self.command = command.kernel
        else:
            self.command = np.asarray(command)

        self.dt = dt
        self.time_supp = np.arange(
            0, self.duration - 0.5 * self.dt, self.dt
        )

    def __str__(self):
        """Return string representation of ArrayStimulus."""
        return "ArrayStimulus of size {:.1f}ms x {} sweeps".format(
            self.duration, self.no_sweeps
        )


class ConvolvedStimulus(BaseStimulus):
    """Stimulus defined by a kernel convolved with a basis.

    See documentation of BaseStimulus for more information about Stimulus
    objects.

    """

    # Methods that must be implemented by derived classes.
    def __init__(
        self, loc, kernel, basis=None, dt=0.1, kernel_max_len=None, label=None
    ):
        """Initialize ConvolvedStimulus.

        Arguments
        ---------
        loc: float
            Constant offset to apply to generated stimulus. Equivalent to
            adding a constant to the generated stimulus.
        kernel: StimulusKernel
            Kernel to convolve with basis. Example: for a train of synaptic-
            like currents, you could use a BiexponentialSynapticKernel to
            specify the shape of a unitary synaptic current.
        basis: 1D array, Stimulus, or StimulusKernel
            Basis with which to convolve kernel. Stimulus command will have the
            same shape as `basis`. Example: for a 100 timestep train of
            synaptic-like currents, this would be a vector of length 100 that
            is 1.0 when a synaptic-like current begins and 0.0 otherwise.
            (Note: for generating synaptic trains, consider using
            PoissonProcess for a simpler interface.)
        dt: float, default 0.1
            Timestep of stimulus in ms.
        kernel_max_len: int or None
            Length at which to truncate the discrete StimulusKernel before
            convolution. Use only if you are using a basis with many timesteps
            and the convolution takes too long to compute.
        label: str
            Descriptive label for stimulus instance.

        """
        self.label = label
        self.dt = dt

        self.loc = loc
        self.kernel = kernel
        self.basis = basis

        if basis is not None:
            self.generate(basis, dt, kernel_max_len)

    def __repr__(self):
        """Return repr(self)."""
        reprstr = (
            'ez.stimtools.ConvolvedStimulus('
            'loc={loc}, '
            'kernel={kernel}, '
            'basis={basis}, '
            'dt={dt}, '
            'label={label}'
            ')'.format(
                loc=self.loc,
                kernel=repr(self.kernel),
                basis=repr(self.basis),
                dt=self.dt,
                label=self.label
            )
        )
        return reprstr

    # Methods that should not be reimplemented by derived classes.
    def generate(self, basis, dt, kernel_max_len=None):
        """Generate stimulus vector by convolving basis with attached kernel.

        Uses an efficient algorithm for sparse convolution. For convolution
        with a sparse signal with a constant offset (e.g., a synaptic train
        with a baseline offset), it is more efficient to specify the offset
        using the `loc` parameter at ConvolvedStimulus initialization than to
        add the offset to `basis`.

        Arguments
        ---------
        basis: 1D array, Stimulus, or StimulusKernel
            Basis with which to convolve kernel. Stimulus command will have the
            same shape as `basis`. Example: for a 100 timestep train of
            synaptic-like currents, this would be a vector of length 100 that
            is 1.0 when a synaptic-like current begins and 0.0 otherwise.
            (Note: for generating synaptic trains, consider using
            PoissonProcess for a simpler interface.)
        dt: float, default 0.1
            Timestep of the stimulus in ms.
        kernel_max_len: int or None
            Length at which to truncate the discrete StimulusKernel before
            convolution. Use only if you are using a basis with many timesteps
            and the convolution takes too long to compute.

        Result
        ------
        Initializes `command` attribute with result of discrete convolution
        of `kernel` attribute with `basis`.

        Side-effects
        ------------
        Initializes `basis` and `time_supp` attributes. Calls `generate` on
        attached `kernel`.

        """
        self.basis = ArrayStimulus(basis, dt)
        self.time_supp = np.arange(0, self.basis.duration - 0.5 * dt, dt)

        # Compute length of kernel to generate.
        if kernel_max_len is None:
            kernel_duration = self.basis.duration
        else:
            kernel_duration = min(kernel_max_len * dt, self.basis.duration)

        # Generate kernel.
        self.kernel.generate(
            duration=kernel_duration, dt=dt, front_padded=False
        )

        # Run discrete convolution.
        output = np.zeros_like(self.basis.command, dtype=np.float64)
        output = self._sparse_convolve(
            self.kernel.kernel.astype(np.float64),
            np.atleast_2d(self.basis.command).astype(np.float64),
            np.atleast_2d(output)
        )
        output += self.loc  # Add offset.

        self.command = output

    @staticmethod
    @nb.jit(
        nb.float64[:, :](nb.float64[:], nb.float64[:, :], nb.float64[:, :]),
        nopython=True
    )
    def _sparse_convolve(kernel, basis, output):
        """Discrete convolution of kernel and basis vector.

        Arguments
        ---------
        kernel : list-like
            Non-sparse stimulus kernel.
        basis : array-like
            Sparse array with which kernel is convolved.
        output : array-like
            Pre-allocated array to hold output.

        Returns
        -------
        output : array-like
            Array containing the convolved stimulus. Same shape as basis.

        """
        for i in range(basis.shape[0]):
            for j in range(basis.shape[1]):
                if basis[i, j] == 0.:
                    continue
                else:
                    end_ind = min(len(kernel), output.shape[1] - j)
                    output[i, j:j + end_ind] += basis[i, j] * kernel[:end_ind]
        return output


# KERNELS

class StimulusKernel(object):
    """Abstract base class of StimulusKernels for use by ConvolvedStimulus.

    Similar to Stimulus objects, but uses a `kernel` rather than `command`
    attribute to store the vector representation of the object.

    """

    def __init__(self):
        raise NotImplementedError

    @property
    def no_timesteps(self):
        """Number of kernel timesteps."""
        if not hasattr(self, 'kernel'):
            return 0
        elif self.kernel.ndim == 1:
            return len(self.kernel)
        else:
            raise NotImplementedError(
                '`no_timesteps` not implemented for non-vector kernel.'
            )

    def generate(self):
        raise NotImplementedError

    def plot(self, ax=None, **pltargs):
        # Check that kernel to plot actually exists.
        if not (hasattr(self, 'kernel') and hasattr(self, 'time_supp')):
            raise AttributeError('`<type>Kernel.kernel` must be initialized '
                                 'by calling `<type>Kernel.generate()` '
                                 'before it can be plotted.')

        # Get default mpl axes.
        if ax is None:
            ax = plt.gca()

        # Make plot.
        ax.plot(self.time_supp, self.kernel, **pltargs)
        ax.set_xlabel('Time (ms)')
        ax.set_ylabel('Amplitude')

        return ax


class ArrayKernel(StimulusKernel):
    """Stimulus kernel constructed from an arbitrary array.

    Analogous to ArrayStimulus.

    See StimulusKernel documentation for for more information about kernels.

    """

    def __init__(self, kernel, dt=0.1, label=None):
        """Initialize ArrayKernel.

        Arguments
        ---------
        kernel: 1D array
            Array to use for kernel.
        dt: float, default 0.1
            Timestep of kernel in ms.
        label: str
            Descriptive label.

        """
        self.label = label

        if issubclass(type(kernel), BaseStimulus):
            self.kernel = kernel.command
        elif issubclass(type(kernel), StimulusKernel):
            self.kernel = kernel.kernel
        else:
            self.kernel = np.asarray(kernel)

        self.dt = dt
        self.duration = self.no_timesteps * self.dt
        self.time_supp = np.arange(
            0, self.duration - 0.5 * self.dt, self.dt
        )

    def __str__(self):
        """Return string representation of ArrayKernel."""
        return "ArrayKernel of duration {:.1f}ms".format(self.duration)

    def generate(self, duration, dt, front_padded=False):
        """Return the ArrayKernel.

        Arguments are only for compatibility with other StimulusKernels and are
        ignored.

        """
        return self.kernel


class BiexponentialSynapticKernel(StimulusKernel):
    """Synaptic kernel with exponential rise and decay.

    See StimulusKernel documentation for more information about kernels.

    """

    def __init__(
        self, size, tau_rise, tau_decay, size_method='amplitude',
        duration=None, dt=0.1, front_padded=False, label=None
    ):
        """Initialize BiexponentialSynapticKernel.

        Arguments
        ---------
        size: float
            Size of waveform. Usually amplitude, but see `size_method`.
        tau_rise, tau_decay: float
            Time constants of mono-exponential rise and decay of the synaptic-
            like current in ms.
        size_method: str, `amplitude` or `AUC`
            Interpret `size` argument as amplitude or area under the curve
            (AUC).
        duration: float
            Duration of stimulus kernel in ms.
        dt: float
            Timestep of kernel in ms.
        front_padded: bool, default False
            Zero-pad the front of the kernel so the onset of the synaptic
            waveform is centered.
        label: str
            Descriptive label.

        """
        self.label = label

        # Store kernel parameters.
        self.size = size
        self.tau_rise = tau_rise
        self.tau_decay = tau_decay
        self.size_method = size_method

        # Generate kernel if optional time params are given.
        if all([x is not None for x in [duration, dt]]):
            self.generate(duration, dt, front_padded)

    def __repr__(self):
        """Return repr(self)."""
        reprstr = (
            'ez.stimtools.BiexponentialSynapticKernel('
            'size={size}, '
            'tau_rise={tau_rise}, '
            'tau_decay={tau_decay}, '
            'size_method={size_method}, '
            'label={label}'
            ')'.format(
                size=self.size, tau_rise=self.tau_rise,
                tau_decay=self.tau_decay, size_method=self.size_method,
                label=self.label
            )
        )
        return reprstr

    def generate(self, duration, dt, front_padded=False):
        """Generate BiexponentialSynapticKernel vector.

        Arguments
        ---------
        duration: float
            Duration of stimulus kernel in ms.
        dt: float
            Timestep of kernel in ms.
        front_padded: bool, default False
            Zero-pad the front of the kernel so the onset of the synaptic
            waveform is centered.

        Result
        ------
        Initializes `kernel` attribute with a biexponential synaptic current-
        like waveform.

        Side-effects
        ------------
        Initializes `time_supp` and `dt` attributes.

        """
        self.time_supp = np.arange(0, duration - 0.5 * dt, dt)

        waveform = (
            np.exp(-self.time_supp / self.tau_decay)
            - np.exp(-self.time_supp / self.tau_rise)
        )
        if self.size_method == 'amplitude':
            waveform /= np.max(waveform)
            waveform *= self.size
        elif self.size_method.upper() == 'AUC':
            waveform *= self.size / (self.tau_decay - self.tau_rise)
        else:
            raise ValueError(
                'Expected `size_method` to be `amplitude` or `AUC`, got {} '
                'instead.'.format(self.size_method)
            )

        if front_padded:
            # Pad with zeros to center kernel.
            waveform = waveform[:(len(self.time_supp) // 2)]
            waveform = np.concatenate(
                [np.zeros(len(waveform) + len(self.time_supp) % 2), waveform]
            )

        self.kernel = waveform
        self.dt = dt

        return waveform


class MonoexponentialSynapticKernel(StimulusKernel):
    """Synaptic kernel with exponential decay.

    See StimulusKernel documentation for more information about kernels.

    """

    def __init__(
        self, size, tau_decay, size_method='amplitude',
        duration=None, dt=0.1, front_padded=False, label=None
    ):
        """Initialize BiexponentialSynapticKernel.

        Arguments
        ---------
        size: float
            Size of waveform. Usually amplitude, but see `size_method`.
        tau_decay: float
            Time constants of mono-exponential decay of the synaptic-like
            current in ms.
        size_method: str, `amplitude` or `AUC`
            Interpret `size` argument as amplitude or area under the curve
            (AUC).
        duration: float
            Duration of stimulus kernel in ms.
        dt: float
            Timestep of kernel in ms.
        front_padded: bool, default False
            Zero-pad the front of the kernel so the onset of the synaptic
            waveform is centered.
        label: str
            Descriptive label.

        """
        self.label = label

        # Store kernel parameters.
        self.size = size
        self.tau_decay = tau_decay
        self.size_method = size_method

        # Generate kernel if optional time params are given.
        if all([x is not None for x in [duration, dt]]):
            self.generate(duration, dt, front_padded)

    def __repr__(self):
        """Return repr(self)."""
        reprstr = (
            'ez.stimtools.MonoexponentialSynapticKernel('
            'size={size}, '
            'tau_decay={tau_decay}, '
            'size_method={size_method}, '
            'label={label}'
            ')'.format(
                size=self.size, tau_decay=self.tau_decay,
                size_method=self.size_method,
                label=self.label
            )
        )
        return reprstr

    def generate(self, duration, dt, front_padded=False):
        """Generate MonoexponentialSynapticKernel vector.

        Arguments
        ---------
        duration: float
            Duration of stimulus kernel in ms.
        dt: float
            Timestep of kernel in ms.
        front_padded: bool, default False
            Zero-pad the front of the kernel so the onset of the synaptic
            waveform is centered.

        Result
        ------
        Initializes `kernel` attribute with an exponential synaptic current-
        like waveform.

        Side-effects
        ------------
        Initializes `time_supp` and `dt` attributes.

        """
        self.time_supp = np.arange(0, duration - 0.5 * dt, dt)

        waveform = (
            np.exp(-self.time_supp / self.tau_decay)
        )
        if self.size_method == 'amplitude':
            waveform /= np.max(waveform)
            waveform *= self.size
        elif self.size_method.upper() == 'AUC':
            waveform *= self.size / self.tau_decay
        else:
            raise ValueError(
                'Expected `size_method` to be `amplitude` or `AUC`, got {} '
                'instead.'.format(self.size_method)
            )

        if front_padded:
            # Pad with zeros to center kernel.
            waveform = waveform[:(len(self.time_supp) // 2)]
            waveform = np.concatenate(
                [np.zeros(len(waveform) + len(self.time_supp) % 2), waveform]
            )

        self.kernel = waveform
        self.dt = dt

        return waveform

# POISSON PROCESS
class PoissonProcess(object):
    #TODO: Inherit from some kind of process object?
    def __init__(self, rate, dt=0.1, label=None):
        """Initialize PoissonProcess.

        Arguments
        ---------
        rate: float 1D array or Stimulus
            Rate of Poisson process in Hz.
        dt: float, default 0.1
            Timestep in ms.
        label: str
            Descriptive label.

        """
        self.dt = dt
        self.label = label
        self.rate = self._coerce_to_stimulus(rate)

    def _coerce_to_stimulus(self, x):
        if issubclass(type(x), BaseStimulus):
            if not np.isclose(x.dt, self.dt):
                raise ValueError(
                    'Expected instance dt={} and argument dt={} '
                    'to be equal.'.format(self.dt, x.dt)
                )
            else:
                return x
        else:
            return ArrayStimulus(x, self.dt, 'Rate of Poisson process.')

    def sample(self, no_samples='auto'):
        """Sample from inhomogenous Poisson process.

        Arguments
        ---------
        no_samples : int
            Number of samples to draw from Poisson process.

        Returns
        -------
        Array with same shape as instance rate containing samples from point
        process.

        """
        rate_in_mHz = 1e-3 * self.rate.command  # Must convert rate in Hz (s^-1) to mHz (ms^-1)
        event_probability = 1.0 - np.exp(-rate_in_mHz * self.dt)
        if no_samples == 'auto':
            samples = (
                np.random.uniform(0.0, 1.0, size=event_probability.shape)
                < event_probability
            )
        elif self.no_sweeps == 1:
            samples = (
                np.random.uniform(0.0, 1.0, size=(no_samples, len(event_probability)))
                < event_probability
            )
        else:
            raise ValueError(
                'Argument `no_samples={}` must be `auto` because '
                'instance `no_sweeps={}` is not one.'.format(
                    no_samples,
                    self.no_sweeps
                )
            )
        return samples.astype(np.int8)

    @property
    def no_sweeps(self):
        return self.rate.no_sweeps

    @property
    def no_timesteps(self):
        return self.rate.no_timesteps

    @property
    def duration(self):
        return self.rate.duration

# SIMPLE STIMULI

class SimpleStimulus(BaseStimulus):
    """Abstract base class for simple stimuli."""

    def __init__(self, loc, ampli, reqd_args, duration, dt, label=None):
        """Initialize SimpleStimulus.

        Must be implemented by derived classes.

        Suggested implementation
        ------------------------
        1. Store loc, ampli, and stimulus-specific positional arguments in
           attributes.
        2. If duration and dt are not None, call generate method.

        """
        raise NotImplementedError(
            'Initialization must be implemented by derived stimulus classes.'
        )

    def generate(self, duration, dt=0.1):
        """Generate stimulus vector.

        Required call signature
        -----------------------
        SimpleStimulus.generate(duration, dt=0.1)

        """
        raise NotImplementedError(
            '`generate` must be implemented by derived stimulus classes.'
        )


class OUStimulus(SimpleStimulus):
    """Ornstein-Uhlenbeck noise stimulus.

    Ornstein-Uhlenbeck (OU) noise is the limiting case of synaptic noise for
    an infinitely large population of synapses with instantaneous rise and the
    same monoexponential decay.

    """

    def __init__(
        self, mean, amplitude, tau, ampli_modulation, mod_period,
        seed=None, duration=None, dt=0.1,
        label=None
    ):
        """Initialize OUStimulus.

        Arguments
        ---------
        mean: float
            Mean of the generative process. Close to the mean of the generated
            signal for very long signals.
        amplitude: float
            Scale of fluctuations.
        tau: float
            Time constant of the OU process.
        ampli_modulation: float >= 0.0
            Fractional amplitude modulation. Set to zero for no amplitude
            modulation.
        mod_period: float > 0.0
            Period of amplitude modulation in ms. Can be zero if
            `ampli_modulation` is also zero.
        seed: int
            Seed value for pseudorandom number generator.
        duration: float
            Duration of OU noise to realize in ms.
        dt: float, default 0.1
            Timestep of signal in ms.
        label: str
            Descriptive label.

        """
        self.label = label

        # Store stimulus parameters.
        self.mean = mean
        self.amplitude = amplitude
        self.tau = tau
        self.ampli_modulation = ampli_modulation
        self.mod_period = mod_period
        self.seed = seed

        # Generate stimulus if optional time params are given.
        if all([x is not None for x in [duration, dt]]):
            self.generate(duration, dt)

    def __repr__(self):
        """Return repr(self)."""
        reprstr = (
            'ez.stimtools.OUStimulus('
            'mean={mean}, '
            'amplitude={amplitude}, '
            'tau={tau}, '
            'ampli_modulation={ampli_modulation}, '
            'mod_period={mod_period}, '
            'seed={seed}, '
            'label={label}'
            ')'.format(
                mean=self.mean,
                amplitude=self.amplitude,
                tau=self.tau,
                ampli_modulation=self.ampli_modulation,
                mod_period=self.mod_period,
                seed=self.seed,
                label=self.label
            )
        )
        return reprstr

    def generate(self, duration, dt):
        """Generate OUStimulus vector.

        Arguments
        ---------
        duration: float
            Duration of signal to realize in ms.
        dt: float
            Timestep of signal to realize in ms.

        Result
        ------
        Initialize `command` attribute with OU noise.

        Side-effects
        ------------
        Initialize `time_supp` and `dt` attributes. Re-seed pseudorandom number
        generator.

        """
        dtype = np.float64  # Datatype to use for realizing noise.
        self.time_supp = np.arange(0, duration - 0.5 * dt, dt)

        # Precompute sinusoidal amplitude modulation.
        if self.ampli_modulation != 0.:
            ampli = self.amplitude * (
                1 + self.ampli_modulation * np.sin(
                    (2 * np.pi / self.mod_period) * self.time_supp
                )
            )
        else:
            ampli = self.amplitude * np.ones_like(self.time_supp)

        # Sample noise.
        np.random.seed(self.seed)
        rands = np.random.standard_normal(len(self.time_supp))

        # Leakily integrate random walk to get Ornstein-Uhlenbeck noise.
        output_ = np.empty_like(self.time_supp).astype(dtype)
        noise = self._integrate_walk(
            output_.astype(dtype), rands.astype(dtype), ampli.astype(dtype),
            dtype(self.mean), dtype(self.tau), dtype(dt)
        )

        # Assign attributes.
        self.command = noise
        self.dt = dt

    @staticmethod
    @nb.jit(
        nb.float64[:](
            nb.float64[:], nb.float64[:], nb.float64[:],
            nb.float64, nb.float64, nb.float64
        ),
        nopython=True
    )
    def _integrate_walk(output_, rands, amplitude, mean, tau, dt):
        """Leakily integrate random walk."""
        output_[0] = mean
        for t in range(1, len(output_)):
            adaptive_term = mean - output_[t - 1]
            random_term = (
                np.sqrt(2 * amplitude[t - 1]**2 * dt / tau)
                * rands[t - 1]
            )
            doutput_ = adaptive_term * dt / tau + random_term
            output_[t] = output_[t - 1] + doutput_

        return output_


class SinStimulus(SimpleStimulus):
    """Sinusoidal stimulus."""

    def __init__(
        self, mean, amplitude, frequency,
        duration=None, dt=0.1,
        label=None
    ):
        """Initialize SinStimulus.

        Arguments
        ---------
        mean: float
            Mean of oscillations.
        amplitude: float
            Amplitude of oscillations.
        frequency: float
            Frequency of oscillations in Hz.
        duration: float
            Duration of stimulus in ms.
        dt: float, default 0.1
            Timestep of stimulus in ms.
        label: str
            Descriptive label.

        """
        self.label = label

        # Store stimulus parameters.
        self.mean = mean
        self.amplitude = amplitude
        self.frequency = frequency

        # Generate stimulus if optional time params are given.
        if all([x is not None for x in [duration, dt]]):
            self.generate(duration, dt)

    def __repr__(self):
        """Return repr(self)."""
        reprstr = (
            'ez.stimtools.SinStimulus('
            'mean={mean}, '
            'amplitude={amplitude}, '
            'frequency={frequency}, '
            'label={label}'
            ')'.format(
                mean=self.mean,
                amplitude=self.amplitude,
                frequency=self.frequency,
                label=self.label
            )
        )
        return reprstr

    def generate(self, duration, dt):
        """Generate SinStimulus vector.

        Arguments
        ---------
        duration: float
            Duration of stimulus in ms.
        dt: float
            Timestep of stimulus in ms.

        Result
        ------
        Initializes `command` attribute with sinusoidal oscillations.

        Side-effects
        ------------
        Initializes `time_supp` and `dt` attributes.

        """
        self.time_supp = np.arange(0, duration - 0.5 * dt, dt)

        # Generate sine wave.
        wave = self.mean + self.amplitude * np.sin(
            2 * np.pi * self.frequency * 1e-3 * self.time_supp  # Convert to Hz
        )

        # Assign attributes.
        self.command = wave
        self.dt = dt


class CosStimulus(SimpleStimulus):
    """Cosinusoidal stimulus."""

    def __init__(
        self, mean, amplitude, frequency,
        duration=None, dt=0.1,
        label=None
    ):
        """Initialize CosStimulus.

        Arguments
        ---------
        mean: float
            Mean of oscillations.
        amplitude: float
            Amplitude of oscillations.
        frequency: float
            Frequency of oscillations in Hz.
        duration: float
            Duration of stimulus in ms.
        dt: float, default 0.1
            Timestep of stimulus in ms.
        label: str
            Descriptive label.

        """
        self.label = label

        # Store stimulus parameters.
        self.mean = mean
        self.amplitude = amplitude
        self.frequency = frequency

        # Generate stimulus if optional time params are given.
        if all([x is not None for x in [duration, dt]]):
            self.generate(duration, dt)

    def __repr__(self):
        """Return repr(self)."""
        reprstr = (
            'ez.stimtools.CosStimulus('
            'mean={mean}, '
            'amplitude={amplitude}, '
            'frequency={frequency}, '
            'label={label}'
            ')'.format(
                mean=self.mean,
                amplitude=self.amplitude,
                frequency=self.frequency,
                label=self.label
            )
        )
        return reprstr

    def generate(self, duration, dt):
        """Generate CosStimulus vector.

        Arguments
        ---------
        duration: float
            Duration of stimulus in ms.
        dt: float
            Timestep of stimulus in ms.

        Result
        ------
        Initializes `command` attribute with cosinusoidal oscillations.

        Side-effects
        ------------
        Initializes `time_supp` and `dt` attributes.

        """
        self.time_supp = np.arange(0, duration - 0.5 * dt, dt)

        # Generate cosine wave.
        wave = self.mean + self.amplitude * np.cos(
            2 * np.pi * self.frequency * 1e-3 * self.time_supp  # Convert to Hz
        )

        # Assign attributes.
        self.command = wave
        self.dt = dt


class SquareWaveStimulus(SimpleStimulus):
    """Square wave stimulus."""

    def __init__(
        self, mean, amplitude, frequency,
        duration=None, dt=0.1,
        label=None
    ):
        """Initialize SquareWaveStimulus.

        Arguments
        ---------
        mean: float
            Mean of oscillations.
        amplitude: float
            Amplitude of oscillations.
        frequency: float
            Frequency of oscillations in Hz.
        duration: float
            Duration of stimulus in ms.
        dt: float, default 0.1
            Timestep of stimulus in ms.
        label: str
            Descriptive label.

        """
        self.label = label

        # Store stimulus parameters.
        self.mean = mean
        self.amplitude = amplitude
        self.frequency = frequency

        # Generate stimulus if optional time params are given.
        if all([x is not None for x in [duration, dt]]):
            self.generate(duration, dt)

    def __repr__(self):
        """Return repr(self)."""
        reprstr = (
            'ez.stimtools.SquareWaveStimulus('
            'mean={mean}, '
            'amplitude={amplitude}, '
            'frequency={frequency}, '
            'label={label}'
            ')'.format(
                mean=self.mean,
                amplitude=self.amplitude,
                frequency=self.frequency,
                label=self.label
            )
        )
        return reprstr

    def generate(self, duration, dt):
        """Generate SquareWaveStimulus vector.

        Arguments
        ---------
        duration: float
            Duration of stimulus in ms.
        dt: float
            Timestep of stimulus in ms.

        Result
        ------
        Initializes `command` attribute with cosinusoidal oscillations.

        Side-effects
        ------------
        Initializes `time_supp` and `dt` attributes.

        """
        self.time_supp = np.arange(0, duration - 0.5 * dt, dt)

        period = 1e3 / self.frequency
        num_full_alternations = int(len(self.time_supp) // (period / dt))
        trailing_duration = duration % period

        # Implementation note: square wave will be constructed with
        # StepStimulus. The duration of each step will be one half-wave.
        durations = [period / 2. for i in range(2 * num_full_alternations)]
        if trailing_duration < period / 2.:
            durations.append(trailing_duration)
        else:
            durations.append(period / 2.)
            durations.append(trailing_duration % (period / 2.))

        amplitudes = [self.amplitude, -self.amplitude] * num_full_alternations
        amplitudes.append(self.amplitude)  # For trailing_duration.
        if trailing_duration >= period / 2.:
            amplitudes.append(-self.amplitude)

        wave = StepStimulus(durations, amplitudes, dt=dt).command + self.mean

        assert len(wave) == len(self.time_supp), "Length of wave {} does not match time_supp {}".format(len(wave), len(self.time_supp))

        # Assign attributes.
        self.command = wave
        self.dt = dt


class ChirpStimulus(SimpleStimulus):
    """Sine wave stimulus of exponentially changing frequency."""

    def __init__(
        self, mean, amplitude, initial_frequency, final_frequency,
        duration=None, dt=0.1,
        label=None
    ):
        """Initialize ChirpStimulus.

        Arguments
        ---------
        mean: float
            Mean of oscillations.
        amplitude: float
            Amplitude of oscillations.
        initial_frequency, final_frequency: float > 0.0
            Initial and final frequencies of oscillations in Hz.
        duration: float
            Duration of oscillations in ms.
        dt: float
            Timestep of stimulus in ms.
        label: str
            Descriptive label.

        """
        # Input checks.
        if np.isclose(initial_frequency, final_frequency):
            warnings.warn(
                '`initial_frequency` and `final_frequency` should '
                'not be identical. Change one of these values or use '
                '`ez.stimtools.SinStimulus` instead.'
            )

        self.label = label

        # Store stimulus parameters.
        self.mean = mean
        self.amplitude = amplitude
        self.initial_frequency = initial_frequency
        self.final_frequency = final_frequency

        # Generate stimulus if optional time params are given.
        if all([x is not None for x in [duration, dt]]):
            self.generate(duration, dt)

    def __repr__(self):
        """Return repr(self)."""
        reprstr = (
            'ez.stimtools.ChirpStimulus('
            'mean={mean}, '
            'amplitude={amplitude}, '
            'initial_frequency={initial_frequency}, '
            'final_frequency={final_frequency}, '
            'label={label}'
            ')'.format(
                mean=self.mean,
                amplitude=self.amplitude,
                initial_frequency=self.initial_frequency,
                final_frequency=self.final_frequency,
                label=self.label
            )
        )
        return reprstr

    def generate(self, duration, dt):
        """Generate ChirpStimulus vector.

        Arguments
        ---------
        duration: float
            Duration of stimulus in ms.
        dt: float
            Timestep of stimulus in ms.

        Result
        ------
        Initializes `command` attribute with chirp.

        Side-effects
        ------------
        Initializes `time_supp` and `dt` attributes.

        """
        self.time_supp = np.arange(0, duration - 0.5 * dt, dt)

        # Generate sine wave.
        freq = np.logspace(
            np.log10(self.initial_frequency),
            np.log10(self.final_frequency),
            endpoint=True,
            num=len(self.time_supp),
            base=10,
        )
        wave = self.mean + self.amplitude * np.sin(
            2 * np.pi * freq * 1e-3 * self.time_supp  # Convert freq to Hz.
        )

        # Assign attributes.
        self.command = wave
        self.dt = dt


# STEP SIMULUS
class StepStimulus(BaseStimulus):
    """Stimulus consisting of square steps.

    See documentation of BaseStimulus for more information about stimulus
    objects.

    """
    def __init__(self, durations, amplitudes, dt=0.1, label=None):
        """Initialize StepStimulus.

        Arguments
        ---------
        durations: float 1D array
            Duration of each step. Must be of same length as `amplitudes`.
        amplitudes: float 1D array
            Amplitude of each step. Must be of same length as `durations`.
        dt: float, default 0.1
            Timestep of signal.
        label:
            Descriptive label.

        """
        if len(durations) != len(amplitudes):
            raise ValueError(
                'Expected lengths of durations and amplitudes to be equal; '
                'got {} and {} instead.'.format(
                    len(durations), len(amplitudes)
                )
            )

        self.dt = dt
        self.durations = durations
        self.amplitudes = amplitudes

        self.label = label

        self._generate()

    def __repr__(self):
        """Return repr(self)."""
        reprstr = (
            'ez.stimtools.StepStimulus('
            'durations={durations}, '
            'amplitudes={amplitudes}, '
            'dt={dt}, '
            'label={label}, '
            ')'.format(
                durations=self.durations,
                amplitudes=self.amplitudes,
                dt=self.dt,
                label=self.label,
            )
        )
        return reprstr

    def _generate(self):
        command_tmp = []
        for dur, ampli in zip(self.durations, self.amplitudes):
            command_tmp.append(ampli * np.ones(int(dur / self.dt)))
        self.command = np.concatenate(command_tmp, axis=-1)
        self.time_supp = np.arange(0, self.duration - 0.5 * self.dt, self.dt)



# COMPOUND STIMULI

class CompoundStimulus(BaseStimulus):
    """Stimulus constructed from other stimuli.

    Generally the result of adding or concatenating other Stimulus objects.
    Not intended for direct use.

    See documentation of BaseStimulus for more information about Stimulus
    objects.

    Attributes
    ----------
    recipe: str
        Recipe for generating CompoundStimulus from its parts. Not necessarily
        callable.
    command, time_supp, no_sweeps, no_timesteps, duration, dt, label
        Same as BaseStimulus.

    """

    def __init__(self, stimulus=None, dt=0.1, label=None):
        """Initialize CompoundStimulus.

        Arguments
        ---------
        stimulus: Stimulus or array
            Stimulus to coerce to a CompoundStimulus. Copies command, recipe,
            time_supp, and dt if possible.
        dt: float, default 0.1
            Timestep of stimulus. Can be overridden by `stimulus`.
        label: str
            Descriptive label.

        """
        self.label = label
        self.dt = dt

        if stimulus is None:
            self.recipe = ''
        elif issubclass(type(stimulus), BaseStimulus):
            if isinstance(stimulus, CompoundStimulus):
                self.recipe = stimulus.recipe
            else:
                self.recipe = repr(stimulus)

            # If stimulus is a single Stimulus object, just copy attributes.
            self.command = stimulus.command
            self.time_supp = stimulus.time_supp
            if stimulus.dt != dt:
                warnings.warn(
                    'stimulus.dt = {} not equal to argument dt = {}; using '
                    'dt from stimulus.'.format(stimulus.dt, dt)
                )
            self.dt = stimulus.dt
        else:
            self.recipe = 'array_like'

            # Assume stimulus is array_like.
            self.command = np.asarray(stimulus)
            self.time_supp = np.arange(
                0,
                self.duration - 0.5 * self.dt,
                self.dt
            )
            assert len(self.time_supp) == self.no_timesteps

    def __str__(self):
        """Return recipe."""
        return self.recipe

    def __add__(self, x):
        """Add CompoundStimulus and x."""
        # Method to add two Stimulus objects together.
        if issubclass(type(x), BaseStimulus):
            # Check that command arrays have compatible shapes.
            if not (hasattr(self, 'command') and hasattr(x, 'command')):
                raise AttributeError('`command` must be initialized by '
                                     'calling `generate` before Stimulus '
                                     'objects can be added.')
            if self.no_timesteps != x.no_timesteps:
                raise ValueError('`command` arrays must have same number of '
                                 'time steps.')

            # Check that time attributes match.
            for attr_ in ['duration', 'dt']:
                if not np.isclose(getattr(self, attr_), getattr(x, attr_)):
                    raise ValueError(
                        'Stimulus {} must be equal.'.format(attr_)
                    )

            # Get recipe for x.
            if isinstance(x, CompoundStimulus):
                x_recipe = x.recipe
            else:
                x_recipe = repr(x)

            # Add command arrays.
            newcommand = self.command + x.command

        # Method to add Stimulus to array_like object.
        else:
            x_recipe = 'array_like'
            newcommand = self.command + np.asarray(x)

        newstimulus = CompoundStimulus()
        newstimulus.recipe = '\n+ '.join([self.recipe, x_recipe])
        newstimulus.command = newcommand
        newstimulus.time_supp = self.time_supp
        newstimulus.dt = self.dt

        return newstimulus

    def __sub__(self, x):
        """Subtract x from CompoundStimulus."""
        # Method to add two Stimulus objects together.
        if issubclass(type(x), BaseStimulus):
            # Check that command arrays have compatible shapes.
            if not (hasattr(self, 'command') and hasattr(x, 'command')):
                raise AttributeError('`command` must be initialized by '
                                     'calling `generate` before Stimulus '
                                     'objects can be subtracted.')
            if self.no_timesteps != x.no_timesteps:
                raise ValueError('`command` arrays must have same number of '
                                 'time steps.')

            # Check that time attributes match.
            for attr_ in ['duration', 'dt']:
                if not np.isclose(getattr(self, attr_), getattr(x, attr_)):
                    raise ValueError(
                        'Stimulus {} must be equal.'.format(attr_)
                    )

            # Get recipe for x.
            if isinstance(x, CompoundStimulus):
                x_recipe = x.recipe
            else:
                x_recipe = repr(x)

            # Add command arrays.
            newcommand = self.command - x.command

        # Method to add Stimulus to array_like object.
        else:
            x_recipe = 'array_like'
            newcommand = self.command - np.asarray(x)

        newstimulus = CompoundStimulus()
        newstimulus.recipe = '\n- '.join([self.recipe, x_recipe])
        newstimulus.command = newcommand
        newstimulus.time_supp = self.time_supp
        newstimulus.dt = self.dt

        return newstimulus


def concatenate(stimuli, dt='auto'):
    """Join Stimulus objects and/or arrays together end to end.

    Arguments
    ---------
    stimuli: list of Stimulus objects and/or 1D or 2D array-like
        Stimuli to concatenate.
    dt: float or `auto`
        Timestep of stimulus to be returned. Use `auto` to try to infer
        timestep from arguments.

    Returns
    -------
    CompoundStimulus containing stimuli joined end to end.

    """
    # Infer dt.
    #TODO: extract into private method.
    dts = [
        stimulus.dt for stimulus in stimuli
        if issubclass(type(stimulus), BaseStimulus)
    ]
    if len(dts) == 0:
        if dt == 'auto':
            raise ValueError(
                'stimuli must contain at least one Stimulus object or '
                'dt must be specified.'
            )
        else:
            pass
    elif dt == 'auto' and not all([dt_ == dts[0] for dt_ in dts]):
        raise ValueError('stimulus.dt is not equal for all stimuli.')
    elif dt != 'auto' and not all([dt_ == dt for dt_ in dts]):
        raise ValueError(
            'stimulus.dt is not equal to argument dt for all stimuli.'
        )
    else:
        # Successful auto dt.
        dt = dts[0]
        del dts

    # Concatenate stimuli together.
    if isinstance(stimuli[0], CompoundStimulus):  # Initialize result.
        result = stimuli[0].copy()
    else:
        result = CompoundStimulus(deepcopy(stimuli[0]), dt=dt)
    ingredients = [result.recipe]  # To hold recipies of all stimuli.

    # Concatenate each additional stimulus onto result
    # and build recipe from ingredients.
    for stimulus in stimuli[1:]:
        if not isinstance(stimulus, CompoundStimulus):
            stimulus = CompoundStimulus(stimulus, dt=dt)
        result.command = _concatenate_along_columns(
            result.command,
            stimulus.command
        )
        ingredients.append(stimulus.recipe)

    # Update support vector.
    result.time_supp = np.arange(
        0, result.duration - result.dt * 0.5, result.dt
    )

    # Store result recipe.
    concatenate_arg_str = ',\n'.join(ingredients)
    concatenate_arg_str_indented = '\t' + concatenate_arg_str.replace(
        '\n',
        '\n\t'
    )
    result.recipe = '\n'.join(
        ['concatenate(', concatenate_arg_str_indented, ')']
    )

    return result


def _concatenate_along_columns(a, b):
    """Concatenate along columns, broadcasting if necessary.

    Treat a and b as row vectors or matrices and concatenate along column axis.
    Raise a ValueError if either a or b has more than two dimensions.

    """
    # Check whether a and b are both 1D. If so, output should still be 1D.
    if (np.ndim(a) == 1) and (np.ndim(b) == 1):
        output_as_1D = True
    else:
        output_as_1D = False

    # Concatenation is easier if both a and b have same number of dimensions,
    # so coerce to matrices first.
    a = np.atleast_2d(np.copy(a))
    b = np.atleast_2d(np.copy(b))

    # This function is not valid if a or b is a multidimensional (>2D) array,
    # so raise a ValueError if wrong input is supplied.
    if (a.ndim > 2) or (b.ndim > 2):
        raise ValueError(
            'Expected arguments to be of dimensionality at most 2, got {} '
            'and {} instead.'.format(np.ndim(a), np.ndim(b))
        )

    # Do concatenation.
    if a.shape[0] == b.shape[0]:
        # Shapes match: no broadcasting required.
        result = np.concatenate([a, b], axis=1)
    elif a.shape[0] == 1:
        # Broadcast a.
        result = np.concatenate([np.tile(a, (b.shape[0], 1)), b], axis=1)
    elif b.shape[0] == 1:
        # Broadcast b.
        result = np.concatenate([a, np.tile(b, (a.shape[0], 1))], axis=1)
    else:
        # Shapes do not match and cannot broadcast since neither a nor b is a
        # row vector.
        raise ValueError(
            'Matrices cannot be concatenated with incompatible number of '
            'rows: {} and {}.'.format(a.shape[0], b.shape[0])
        )

    # Coerce output to vector if both a and b were vectors.
    if output_as_1D:
        result = result.flatten()

    return result




"""
Created on Tue Jan 30 12:27:39 2018

@author: Emerson

`Recording` class to hold recs as a np.array with built-in methods for plotting
and test-pulse fitting. Implements a factory class `ABFLoader` for loading recordings in
Axon binary format (.abf).

Compatible with python 2 and 3 as of Feb. 5, 2018.
"""

# IMPORT MODULES

from __future__ import division

from warnings import warn
from copy import deepcopy

import numpy as np
import matplotlib.pyplot as plt
from scipy import optimize
from neo.io import AxonIO


# OBJECTS FOR LOADING RECORDINGS

class BaseRecordingLoader(object):
    """Abstract base class for RecordingLoaders.

    Children implement the following methods needed to load
    electrophysiological recordings of a given format:
    - _read_data_from_file(file_name)
    - _get_sampling_interval_in_ms(file_data)
    - _coerce_to_recording(file_data, sampling_interval)

    These private methods are called by _load_single_file(file_name).

    """
    def __init__(self):
        raise NotImplementedError

    def load(self, file_names):
        """Load recordings from files.

        Arguments
        ---------
        file_names : list or str
            list of files to import, or str for a single file

        Returns
        -------
        List of np.array-like recordings each with dimensionality
        [channels, samples, sweeps].

        """
        if isinstance(file_names, str):
            file_names = [file_names]

        recordings = []
        for file_name in file_names:
            recordings.append(self._load_single_file(file_name))
        return recordings

    def _load_single_file(self, file_name):
        file_data = self._read_data_from_file(file_name)
        sampling_intervals = self._get_sampling_intervals_in_ms(file_data)
        assert all(np.isclose(sampling_intervals, sampling_intervals[0]))
        recording = self._coerce_to_recording(file_data, sampling_intervals[0])
        return recording

    def _read_data_from_file(self, file_name):
        raise NotImplementedError

    def _get_sampling_intervals_in_ms(self, file_data):
        raise NotImplementedError

    def _coerce_to_recording(self, file_data, sampling_interval):
        raise NotImplementedError


class ABFLoader(BaseRecordingLoader):
    """Load recordings in Axon binary format (.abf).

    Recordings are loaded by passing a list of file names to the `load()`
    method.

    """
    def __init__(self):
        pass

    def _read_data_from_file(self, file_name):
        return AxonIO(file_name).read()[0].segments

    def _get_sampling_intervals_in_ms(self, file_data):
        sampling_intervals = []
        for sweep in file_data:
            for signal in sweep.analogsignals:
                sampling_intervals.append(1e3 / signal.sampling_rate.item())
        return sampling_intervals

    def _coerce_to_recording(self, file_data, sampling_interval):
        sweeps = file_data

        no_channels = len(sweeps[0].analogsignals)
        no_samples = len(sweeps[0].analogsignals[0])
        no_sweeps = len(sweeps)

        sweeps_arr = np.empty(
            (no_channels, no_samples, no_sweeps), dtype=np.float64
        )

        # Fill the array one sweep at a time.
        for sweep_ind in range(no_sweeps):

            for chan_ind in range(no_channels):

                signal = sweeps[sweep_ind].analogsignals[chan_ind]
                signal = np.squeeze(signal)

                assert len(signal) == sweeps_arr.shape[1], (
                    'Not all channels are sampled at the same rate.'
                )

                sweeps_arr[chan_ind, :, sweep_ind] = signal

        return Recording(sweeps_arr, dt=sampling_interval)


# PYTHON OBJECT FOR REPRESENTING RECORDINGS

class Recording(np.ndarray):
    """Thin wrapper of numpy.ndarray with add-ons for common ephys tasks.

    Recording objects are 3D arrays with dimensionality [channel, time, sweep].

    Extra attributes
    ----------------
    no_channels: int
        Number of channels in the recording (e.g., current, voltage, etc).
    no_sweeps, no_timesteps: int
        Number of sweeps, timesteps in the recording.
    duration: float
        Duration of one sweep in ms.
    dt: float
        Sampling interval in ms. 1000.0/sampling rate in Hz.
    time_supp: float 1D array
        Time support vector for one sweep of the recording in ms. Always starts
        at zero.

    Extra methods
    -------------
    plot
    fit_test_pulse

    """

    def __new__(cls, input_array, dt=0.1):
        """Instantiate new Recording given an array of data.

        Allows new Recording objects to be created using np.array-type syntax;
        i.e., by passing Recording a nested list or existing np.array.
        """
        if np.ndim(input_array) != 3:
            raise ValueError(
                'Expected `input_array` ndim == 3, got {} instead. '
                'Dimensionality must be `[channel, time, sweep]`.'.format(
                    np.ndim(input_array)
                )
            )

        # Convert input_array to a np.ndarray, and subsequently to a Recording.
        obj = np.asarray(input_array).view(cls)

        # Initialize attributes.
        obj.dt = dt

        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return None

        # Copy time step.
        self.dt = getattr(obj, 'dt', None)

    @property
    def no_channels(self):
        """Number of channels in instance."""
        assert self.ndim == 3
        return self.shape[0]

    @property
    def no_sweeps(self):
        """Number of sweeps in instance."""
        assert self.ndim == 3
        return self.shape[2]

    @property
    def no_timesteps(self):
        """Number of sweeps in instance."""
        assert self.ndim == 3
        return self.shape[1]

    @property
    def duration(self):
        """Duration of one sweep in ms."""
        return self.no_timesteps * self.dt

    @property
    def time_supp(self):
        """Time support vector for one sweep.

        Gives time from start of sweep in ms. Always starts at zero.

        """
        if getattr(self, '_time_supp', None) is None:
            self._init_time_supp()
        return self._time_supp

    def _init_time_supp(self):
        self._time_supp = np.arange(0, self.duration - 0.5 * self.dt, self.dt)

    def plot(self, single_sweep=False, downsample=10):
        """Quick inspection of Recording.

        Note that x-axis values correspond to inds of the time axis of the array.
        """
        ### Check for correct input ###

        # Check single_sweep.
        if not isinstance(single_sweep, bool):
            raise TypeError('`single_sweep` must be bool.')

        # Check downsample.
        if downsample is None:
            downsample = 1

        elif not isinstance(downsample, int):
            raise TypeError('`downsample` must be int or None.')

        elif downsample < 1:
            raise ValueError('`downsample` must be an int > 0. or None.')

        ### Select data to plot ###
        if not single_sweep:
            plotting_data = self
        else:
            plotting_data = self[:, :, 0][:, :, np.newaxis]

        ### Make plot ###
        # Preserves indexes.
        x_vector = np.arange(0, self.shape[1], downsample)
        plt.figure(figsize=(10, 7))

        for i in range(self.shape[0]):

            # Force all subplots to share x-axis.
            if i == 0:
                ax0 = plt.subplot(self.shape[0], 1, 1)
            else:
                plt.subplot(self.shape[0], 1, i + 1, sharex=ax0)

            plt.title('Channel {}'.format(i))
            plt.plot(x_vector, plotting_data[i, ::downsample, :],
                     'k-',
                     linewidth=0.5)
            plt.xlabel('Time (timesteps)')

        plt.tight_layout()
        plt.show()

    def fit_test_pulse(self, baseline, steady_state, **kwargs):
        """Extract R_input and (optionally) R_a from test pulse.

        `baseline` and `steady_state` should be passed tuples of indexes over
        which to take measurements on each sweep.

        Set `verbose` to False to prevent printing results.

        tau: 3 tuple, optional
        --  Tuple of test pulse start and range over which to calculate tau in *indexes*.
        plot_tau: bool, default False
        --  Optionally plot the tau fit.
        """
        ### Inputs ###

        # Set kwarg defaults.
        kwargs.setdefault('V_chan', 1)
        kwargs.setdefault('I_chan', 0)
        kwargs.setdefault('V_clamp', True)
        kwargs.setdefault('verbose', True)
        kwargs.setdefault('tau', None)
        kwargs.setdefault('plot_tau', False)

        # Check for correct inputs.
        if not isinstance(baseline, tuple):
            raise TypeError('Expected type tuple for `baseline`; got {} '
                            'instead.'.format(type(baseline)))
        elif any([not isinstance(entry, int) for entry in baseline]):
            raise TypeError('Expected tuple of ints for `baseline`.')
        elif len(baseline) != 2:
            raise TypeError('Expected tuple of len 2 specifying start and '
                            'stop positions for `baseline`.')
        elif any([entry > self.shape[1] for entry in baseline]):
            raise ValueError('`baseline` selection out of bounds for channel '
                             'of length {}.'.format(self.shape[1]))

        if not isinstance(steady_state, tuple):
            raise TypeError('Expected type tuple for `steady_state`; got {} '
                            'instead.'.format(type(steady_state)))
        elif any([not isinstance(entry, int) for entry in steady_state]):
            raise TypeError('Expected tuple of ints for `steady_state`.')
        elif len(steady_state) != 2:
            raise TypeError('Expected tuple of len 2 specifying start and '
                            'stop positions for `steady_state`.')
        elif any([entry > self.shape[1] for entry in steady_state]):
            raise ValueError('`steady_state` selection out of bounds for '
                             'channel of length {}.'.format(self.shape[1]))

        if steady_state[0] < baseline[1]:
            raise ValueError('Steady state measurement must be taken after '
                             ' end of baseline.')

        if not isinstance(kwargs['V_clamp'], bool):
            raise TypeError('Expected `V_clamp` to be type bool; got {} '
                            'instead.'.format(type(kwargs['V_clamp'])))

        if not isinstance(kwargs['verbose'], bool):
            raise TypeError('Expected `verbose` to be type bool; got {} '
                            'instead.'.format(type(kwargs['verbose'])))

        ### Main ###

        # Create dict to hold output.
        output = {}

        # Calculate R_input.
        V_baseline = self[kwargs['V_chan'], slice(*baseline), :].mean(axis=0)
        I_baseline = self[kwargs['I_chan'], slice(*baseline), :].mean(axis=0)
        V_test = self[kwargs['V_chan'], slice(*steady_state), :].mean(axis=0)
        I_test = self[kwargs['I_chan'], slice(*steady_state), :].mean(axis=0)

        delta_V_ss = V_test - V_baseline
        delta_I_ss = I_test - I_baseline

        R_input = 1000 * delta_V_ss / delta_I_ss
        output['R_input'] = R_input

        # Calculate R_a.
        if kwargs['V_clamp']:

            if delta_V_ss.mean() < 0:
                I_peak = self[kwargs['I_chan'],
                              slice(baseline[1], steady_state[0]),
                              :].min(axis=0)
            else:
                I_peak = self[kwargs['I_chan'],
                              slice(baseline[1], steady_state[0]),
                              :].max(axis=0)

            R_a = 1000 * delta_V_ss / (I_peak - I_baseline)
            output['R_a'] = R_a

        if kwargs['tau'] is not None:
            try:
                self.dt
            except NameError:
                raise RuntimeError('dt (timestep) must be set to fit tau')

            if not kwargs['V_clamp']:

                V_copy = deepcopy(self[kwargs['V_chan'], :, :])
                V_copy = V_copy.mean(axis=1)

                pulse_start = kwargs['tau'][0]
                fitting_range = kwargs['tau'][-2:]

                p0 = [V_copy[slice(*baseline)].mean(), V_copy[slice(*steady_state)].mean(), 10]
                p, fitted_pts = self._exponential_optimizer_wrapper(V_copy[slice(*fitting_range)], p0, self.dt)

                output['tau'] = p[2]

                if kwargs['plot_tau']:
                    plt.figure()
                    plt.plot(
                        np.arange(0, (len(V_copy) - 0.5) * self.dt, self.dt), V_copy,
                        'k-', lw=0.5
                    )
                    plt.plot(
                        np.linspace(fitting_range[0] * self.dt, fitting_range[1] * self.dt, fitted_pts.shape[1]),
                        fitted_pts[0, :],
                        'b--'
                    )
                    plt.show()

            else:
                raise NotImplementedError(
                    'Tau fitting for V-clamp is not implemented.')

        # Optionally, print results.
        if kwargs['verbose']:
            print('\n\n### Test-pulse results ###')
            print('R_in: {} +/- {} MOhm'.format(round(R_input.mean(), 1),
                                                round(R_input.std())))

            if kwargs['V_clamp']:
                print('R_a: {} +/- {} MOhm'.format(round(R_a.mean()),
                                                   round(R_a.std())))

        return output

    def _exponential_curve(self, p, t):
        """Three parameter exponential.

        I = (A + C) * exp (-t/tau) + C

        p = [A, C, tau]
        """

        A = p[0]
        C = p[1]
        tau = p[2]

        return (A + C) * np.exp(-t/tau) + C

    def _compute_residuals(self, p, func, Y, X):
        """Compute residuals of a fitted curve.

        Inputs:
            p       -- vector of function parameters
            func    -- a callable function
            Y       -- real values
            X       -- vector of points on which to compute fitted values

        Returns:
            Array of residuals.
        """

        if len(Y) != len(X):
            raise ValueError('Y and X must be of the same length.')

        Y_hat = func(p, X)

        return Y - Y_hat

    def _exponential_optimizer_wrapper(self, I, p0, dt=0.1):

        t = np.arange(0, len(I) * dt, dt)[:len(I)]

        p = optimize.least_squares(self._compute_residuals, p0, kwargs={
        'func': self._exponential_curve,
        'X': t,
        'Y': I
        })['x']

        no_pts = 500

        fitted_points = np.empty((2, no_pts))
        fitted_points[1, :] = np.linspace(t[0], t[-1], no_pts)
        fitted_points[0, :] = self._exponential_curve(p, fitted_points[1, :])

        return p, fitted_points





"""Unit tests for `stimtools`."""

import unittest

import numpy as np
import numpy.testing as npt

from ezephys import stimtools as st

class TestStepStimulus(unittest.TestCase):
    """Unit tests for `stimtools.StepStimulus` stimulus class."""

    def test_construction(self):
        stepstim = st.StepStimulus([0.25, 0.75], [-1., 1.], dt=0.1)
        npt.assert_array_equal(
            stepstim.command,
            [-1., -1., 1., 1., 1., 1., 1., 1., 1.]
        )


class TestSquareWaveStimulus(unittest.TestCase):

    def test_wave_is_square(self):
        duration = 1000.
        timestep = 0.1
        square_wave = st.SquareWaveStimulus(0.5, 0.5, 1, duration, timestep)
        first_half_one = np.allclose(
            square_wave.command[:int(duration / 2. / timestep)], 1.
        )
        second_half_zero = np.allclose(
            square_wave.command[int(duration / 2. / timestep):], 0.
        )
        self.assertTrue(
            first_half_one and second_half_zero,
            'Square wave is not first half up and second half down.'
        )

    def test_time_supp_and_command_lengths_match(self):
        duration = 2000.
        timestep = 0.1
        square_wave = st.SquareWaveStimulus(0.5, 0.5, 2, duration, timestep)
        self.assertEqual(len(square_wave.command), len(square_wave.time_supp))


class TestCompoundStimulus(unittest.TestCase):
    """Unit tests for `stimtools.CompoundStimulus` stimulus class."""

    def test_add(self):
        """Test `stimtools.CompoundStimulus.__add__` behaviour."""
        dt = 1
        initial_command = np.array([1, 2, 3, 2])
        compound = st.CompoundStimulus(initial_command, dt=dt)

        for const_offset in np.linspace(-10, 10, 7):
            tmp_compound = compound + const_offset
            npt.assert_array_almost_equal(
                tmp_compound.command,
                initial_command + const_offset,
                err_msg='Add constant to `CompoundStimulus` not '
                'equivalent to adding to command.'
            )

        variable_offset = np.array([3, 3, 2, 1])
        tmp_compound = compound + variable_offset
        npt.assert_array_almost_equal(
            tmp_compound.command,
            initial_command + variable_offset,
            err_msg='Add vector to `CompoundStimulus` not '
            'equivalent to adding to command.'
        )

        tmp_compound = compound + st.CompoundStimulus(variable_offset, dt=dt)
        npt.assert_array_almost_equal(
            tmp_compound.command,
            initial_command + variable_offset,
            err_msg='Adding `CompoundStimulus` objects not equivalent to '
            'adding commands.'
        )

    def test_subtract(self):
        """Test `stimtools.CompoundStimulus.__sub__` behaviour."""
        dt = 1
        initial_command = np.array([1, 2, 3, 2])
        compound = st.CompoundStimulus(initial_command, dt=dt)

        for const_offset in np.linspace(-10, 10, 7):
            tmp_compound = compound - const_offset
            npt.assert_array_almost_equal(
                tmp_compound.command,
                initial_command - const_offset,
                err_msg='Subtract constant from `CompoundStimulus` not '
                'equivalent to subtraction from command.'
            )

        variable_offset = np.array([3, 3, 2, 1])
        tmp_compound = compound - variable_offset
        npt.assert_array_almost_equal(
            tmp_compound.command,
            initial_command - variable_offset,
            err_msg='Subtract vector from `CompoundStimulus` not '
            'equivalent to subtraction from command.'
        )

        tmp_compound = compound - st.CompoundStimulus(variable_offset, dt=dt)
        npt.assert_array_almost_equal(
            tmp_compound.command,
            initial_command - variable_offset,
            err_msg='Adding `CompoundStimulus` objs not equivalent to adding '
            'commands.'
        )


class TestConcatenate(unittest.TestCase):
    """Unit tests for `stimtools.concatenate`."""

    def test_CompoundStimulus_only(self):
        """Test concatenation of CompoundStimulus objects."""
        dt = 0.7
        command1 = np.array([1, 2, 3, 2])
        command2 = np.array([1, 1, 2, 2, 3])

        # Test behaviour with good dt.
        compound1 = st.CompoundStimulus(command1, dt=dt)
        compound2 = st.CompoundStimulus(command2, dt=dt)

        new_compound = st.concatenate([compound1, compound2])

        npt.assert_array_almost_equal(
            new_compound.command,
            np.concatenate([command1, command2], axis=-1),
            err_msg='Concatenatation of vector with CompoundStimulus not equivalent to '
            'numpy concatenation.'
        )
        self.assertEqual(
            new_compound.no_timesteps,
            len(command1) + len(command2),
            'no_timesteps does not match length of commands inputs.'
        )
        self.assertEqual(
            new_compound.dt, dt, 'Incorrect dt in concatenated stimulus.'
        )

        # Test behaviour when bad dt is supplied.
        compound1 = st.CompoundStimulus(command1, dt=1)
        compound2 = st.CompoundStimulus(command2, dt=2)
        with self.assertRaises(
            ValueError,
            msg='Value error not raised for mismatched dt using auto dt.'
        ):
            st.concatenate([compound1, compound2])

        compound1 = st.CompoundStimulus(command1, dt=dt)
        compound2 = st.CompoundStimulus(command2, dt=dt)
        with self.assertRaises(
            ValueError,
            msg='Value error not raised for mismatched dt using specified dt.'
        ):
            st.concatenate([compound1, compound2], dt=dt*2)

    def test_vectors_only(self):
        """Test concatenation of numpy arrays using stimtools.concatenate."""
        dt = 0.7
        command1 = np.array([1, 2, 3, 2])
        command2 = np.array([1, 1, 2, 2, 3])

        new_compound = st.concatenate([command1, command2], dt=dt)

        npt.assert_array_almost_equal(
            new_compound.command,
            np.concatenate([command1, command2], axis=-1),
            err_msg='Concatenatation of vector with CompoundStimulus not equivalent to '
            'numpy concatenation.'
        )
        self.assertEqual(
            new_compound.no_timesteps,
            len(command1) + len(command2),
            'no_timesteps does not match length of commands inputs.'
        )
        self.assertEqual(
            new_compound.dt, dt, 'Incorrect dt in concatenated stimulus.'
        )

        # Test behaviour when bad dt is supplied.
        with self.assertRaises(
            ValueError,
            msg='Value error not raised when required dt is not supplied.'
        ):
            st.concatenate([command1, command2])


if __name__ == '__main__':
    unittest.main()



"""Install ezephys."""

import os
import sys
import re
import setuptools

# Read ezephys version number from ezephys/__init__.py
currdir = os.path.dirname(os.path.realpath(__file__))
with open(os.path.join(currdir, 'ezephys', '__init__.py'), 'r') as f:
    version = re.search(r"__version__ = '([^']+)'", f.read()).group(1)
    f.close()

# Set minimal install requirements based on Python version.
install_requires = ['matplotlib', 'seaborn', 'numpy']
if sys.version_info.major == 2:
    # Reasons for pinning packages:
    # - llvmlite versions >=0.32 do not support Python 2.
    # - numba version needs to match llvmlite
    # - quantities versions >= 0.13 do not support Python 2.
    # - neo >= 0.9 do not support Python 2.
    install_requires.extend(
        ['numba==0.43.1', 'llvmlite<0.32.0', 'quantities==0.12.5', 'neo<0.9.0']
    )
elif sys.version_info.major == 3:
    install_requires.extend(['numba', 'llvmlite', 'neo'])
else:
    raise RuntimeError(
        'Python version {} is not supported.'.format(sys.version_info.major)
    )

# Read long description from restructured text file.
with open(os.path.join(currdir, 'long_description.rst'), 'r') as f:
    long_description = f.read()
    f.close()

setuptools.setup(
    name='ezephys',
    version=version,
    description='Tools for working with electrophysiological data.',
    long_description=long_description,
    license='MIT',
    author='Emerson Harkin',
    author_email='emerson@efharkin.com',
    url='https://github.com/efharkin/ez-ephys',
    download_url='https://pypi.org/project/ez-ephys/',
    keywords='neuroscience,electrophysiology',
    packages=['ezephys'],
    install_requires=install_requires,
    python_requires=(
        '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*'
    ),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
)



"""Unit tests for `rectools`."""

import unittest

import numpy as np
import numpy.testing as npt

import ezephys.rectools as rt

class TestRecording(unittest.TestCase):
    """Tests for `rectools.Recording`."""

    def test_type_preservation(self):
        """Test that `Recording` objects are returned from various ops on `Recording` objects."""
        test_rec = rt.Recording([[[0, 1, 2]]])
        self.assertTrue(
            isinstance(test_rec, rt.Recording),
            'Result of `Recording` constructor is not of `Recording` type.',
        )
        self.assertTrue(
            isinstance(test_rec[..., 1:], rt.Recording),
            'Result of slicing `Recording` is not of `Recording` type.',
        )
        self.assertTrue(
            isinstance(test_rec[..., 1], rt.Recording),
            'Result of retrieving single element of `Recording` is not of '
            '`Recording` type.',
        )

    def test_dt_preservation(self):
        """Test that `dt` is preserved after indexing `Recording`."""
        dt = 0.6767335
        test_rec = rt.Recording([[[0, 1, 2]]], dt=dt)
        self.assertEqual(
            test_rec.dt,
            dt,
            'Assigned `dt` not equal to attribute `dt`; test code probably '
            'broken.',
        )
        self.assertEqual(
            test_rec[..., 1:].dt,
            dt,
            '`Recording.dt` attribute altered by slicing.',
        )
        self.assertEqual(
            test_rec[..., 1].dt,
            dt,
            '`dt` attribute altered by retrieving single element of `Recording`.',
        )

    def test_no_timesteps_property(self):
        """Test that no_timesteps reflects number of timesteps in recording."""
        expected_values = {
            'no_timesteps': 1000,
            'no_sweeps': 10,
            'no_channels': 4,
        }
        test_rec = rt.Recording(
            np.zeros(
                [
                    expected_values['no_channels'],
                    expected_values['no_timesteps'],
                    expected_values['no_sweeps'],
                ]
            ),
            dt=0.1,
        )
        self.assertEqual(
            test_rec.no_timesteps,
            expected_values['no_timesteps'],
            'Expected {} for `no_timesteps` property; got {} instead.'.format(
                expected_values['no_timesteps'], test_rec.no_timesteps
            ),
        )

    def test_no_sweeps_property(self):
        """Test that no_sweeps reflects number of sweeps in recording."""
        expected_values = {
            'no_timesteps': 1000,
            'no_sweeps': 10,
            'no_channels': 4,
        }
        test_rec = rt.Recording(
            np.zeros(
                [
                    expected_values['no_channels'],
                    expected_values['no_timesteps'],
                    expected_values['no_sweeps'],
                ]
            ),
            dt=0.1,
        )
        self.assertEqual(
            test_rec.no_sweeps,
            expected_values['no_sweeps'],
            'Expected {} for `no_sweeps` property; got {} instead.'.format(
                expected_values['no_sweeps'], test_rec.no_sweeps
            ),
        )

    def test_no_channels_property(self):
        """Test that no_channels reflects number of channels in recording."""
        expected_values = {
            'no_timesteps': 1000,
            'no_sweeps': 10,
            'no_channels': 4,
        }
        test_rec = rt.Recording(
            np.zeros(
                [
                    expected_values['no_channels'],
                    expected_values['no_timesteps'],
                    expected_values['no_sweeps'],
                ]
            ),
            dt=0.1,
        )
        self.assertEqual(
            test_rec.no_channels,
            expected_values['no_channels'],
            'Expected {} for `no_channels` property; got {} instead.'.format(
                expected_values['no_channels'], test_rec.no_channels
            ),
        )

    def test_duration_property(self):
        """Test that duration reflects number of duration in recording."""
        recording_dt = 0.1
        recording_shape = {
            'no_timesteps': 1000,
            'no_sweeps': 10,
            'no_channels': 4,
        }
        expected_duration = recording_shape['no_timesteps'] * recording_dt
        test_rec = rt.Recording(
            np.zeros(
                [
                    recording_shape['no_channels'],
                    recording_shape['no_timesteps'],
                    recording_shape['no_sweeps'],
                ]
            ),
            dt=recording_dt,
        )
        npt.assert_almost_equal(
            test_rec.duration,
            expected_duration,
            err_msg='Expected {} for `duration` property; got {} instead.'.format(
                expected_duration, test_rec.duration
            ),
        )

    def test_time_supp_length_matches_no_timesteps(self):
        """Ensure time_supp is a valid time support vector for input signal."""
        for no_timesteps in [5, 578, 993, 300072]:
            for dt in [0.1, 0.5, 3.0]:
                test_rec = rt.Recording(np.empty([6, no_timesteps, 1]), dt=dt)
                self.assertEqual(
                    len(test_rec.time_supp),
                    no_timesteps,
                    'Expected length of time_supp {} to match no_timesteps of '
                    'input {}.'.format(len(test_rec.time_supp), no_timesteps),
                )

    def test_time_supp_length_matches_no_timesteps_after_slicing(self):
        test_rec = rt.Recording(np.empty([3, 4387, 2]), dt=0.4)
        test_sliced_rec = test_rec[:, 300:557, :]
        self.assertEqual(
            len(test_sliced_rec.time_supp),
            test_sliced_rec.no_timesteps,
            'Expected length of `time_supp` {} to equal no_timesteps of '
            'Recording {} after slicing.'.format(
                len(test_sliced_rec.time_supp), test_sliced_rec.no_timesteps
            ),
        )

if __name__ == '__main__':
    unittest.main()

"""
# LOAD TEST RECORDING

# Load all ABF files in the example data directory
EX_DATA_DIR = 'test_data'
PATHS = []
for fname in os.listdir(EX_DATA_DIR):
    if fname[-4:].lower() == '.abf':
        PATHS.append(os.path.join(EX_DATA_DIR, fname))
loader = rt.ABFLoader()
test_rec_ls = loader.load(PATHS)

# Assign one of the recordings to test_rec and plot it
test_rec = test_rec_ls[1]
test_rec.plot()


# TEST SUPPORT STRUCTURES

# Set timestep in ms
test_rec.set_dt(0.1)

# Test t_vec
plt.figure()
plt.title('t_vec test')
plt.plot(test_rec.t_vec, test_rec[0, :, 0], 'k-')
plt.xlabel('Time (ms)')
plt.show()

# Test t_mat, which should be the same shape as test_rec
plt.figure()
plt.title('t_mat test')
plt.plot(test_rec.t_mat[0, :, :], test_rec[0, :, :], 'k-', alpha=0.5)
plt.xlabel('Time (ms)')
plt.show()
"""
